/*
 * (C) Copyright 2011
 * CompuLab, Ltd., <www.compulab.co.il>
 * Ilya Ledvich <ilya@compulab.co.il>
 * Igor Grinberg <grinberg@compulab.co.il>
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc.
 */

#include <common.h>
#include <part.h>
#include <fat.h>
#include <mmc.h>

#include <asm/arch/cpu.h>
#include <asm/arch/bits.h>
#include <asm/arch/sys_info.h>
#include <asm/arch/mem.h>

#include "common.h"
#include "prcm.h"
#include "memory.h"
#include "board.h"

/* Table of supported board probe callbacks */
static int (*board_probe_table[])(struct board_data *) = {
	cm_t3530_probe,
	cm_t3730_probe,
	cm_t3517_probe,
};

/* Board data of the recognized board */
static struct board_data bd __attribute__ ((section (".data"))) = {};
struct board_data *board = &bd;


/*****************************************
 * Routine: board_probe
 * Description: Determine board type
 * Return: 0 on success, 1 otherwise
 *****************************************/
static int board_probe(struct board_data *data)
{
	int i;

	for (i = 0; i < ARRAY_SIZE(board_probe_table); i++)
		if ((board_probe_table[i])(data) == 0)
			return 0;

	return 1;
}

/****************************************
 * Routine: board_init
 * Description: Early hardware init.
 ****************************************/
int board_init(void)
{
	led_enable();

	if (board->board_init)
		return board->board_init();

	return 0;
}

#define NAND_UNKNOWN	0
#define NAND_PRESENT	1
#define NAND_ABSENT	2

static int nand_status;

/************************************************************
 * Routine: cm_t3x_get_mem_type(void) - returns the kind of
 * memory we are trying to boot from. Uses SYS_BOOT settings.
 ************************************************************/
static u32 cm_t3x_get_mem_type(void)
{
	u32 mem_type = get_sysboot_value();

	/* MMC Boot button pressed or no NAND available */
	if (mem_type & 0x20 || nand_status == NAND_ABSENT)
		return MMC_NAND;

	/* Any other case - boot from NAND */
	return GPMC_NAND;
}

/************************************************************
 * Routine: get_mem_type(void) - returns the kind of
 * memory we are trying to boot from.
 * Calls the board specific get_mem_type() function if defined.
 *************************************************************/
u32 get_mem_type(void)
{
	if (board->get_mem_type)
		return board->get_mem_type();

	return cm_t3x_get_mem_type();
}

/***********************************************************
 * Routine: s_init
 * Description: Does early system init of muxing and clocks.
 * - Called at time when only stack is available.
 ***********************************************************/
void s_init(void)
{
	int err = board_probe(board);

	if (err)
		hang();

	/* Execute early init functions */
	board->watchdog_init();
	board->try_unlock_memory();
	board->set_muxconf_regs();
	delay(100);
	board->prcm_init();
	board->per_clocks_enable();
	board->setup_memory_controller();
}

/******************************************************
 * Routine: misc_init_r
 ******************************************************/
int misc_init_r(void)
{
	if (board->board_name)
		pr_info("Detected board: %s\n", board->board_name);

	if (board->misc_init_r)
		return board->misc_init_r();

	return 1;
}

int cm_t3x_nand_init(u32 *gpmc_cfg)
{
	u32 mem_type = get_mem_type();

	if (!gpmc_cfg)
		return 0;

	/* global settings */
	__raw_writel(0x10, GPMC_SYSCONFIG);	/* smart idle */
	__raw_writel(0x0, GPMC_IRQENABLE);	/* isr's sources masked */
	__raw_writel(0, GPMC_TIMEOUT_CONTROL);	/* timeout disable */

	/*
	 * Set the GPMC Vals . For NAND boot on CM-T3X, NAND is mapped at CS0.
	 * We configure only GPMC CS0 with required values.
	 * Configuring other devices at other CS in done in U-Boot anyway.
	 * So we don't have to bother doing it here.
	 */
	__raw_writel(0 , GPMC_CONFIG7 + GPMC_CONFIG_CS0);
	delay(1000);

	__raw_writel(gpmc_cfg[0], GPMC_CONFIG1 + GPMC_CONFIG_CS0);
	__raw_writel(gpmc_cfg[1], GPMC_CONFIG2 + GPMC_CONFIG_CS0);
	__raw_writel(gpmc_cfg[2], GPMC_CONFIG3 + GPMC_CONFIG_CS0);
	__raw_writel(gpmc_cfg[3], GPMC_CONFIG4 + GPMC_CONFIG_CS0);
	__raw_writel(gpmc_cfg[4], GPMC_CONFIG5 + GPMC_CONFIG_CS0);
	__raw_writel(gpmc_cfg[5], GPMC_CONFIG6 + GPMC_CONFIG_CS0);
	/* Enable the GPMC Mapping */
	__raw_writel(((OMAP34XX_GPMC_CS0_SIZE & 0xF) << 8) |
		     ((NAND_BASE_ADR >> 24) & 0x3F) | (1 << 6),
		     GPMC_CONFIG7 + GPMC_CONFIG_CS0);
	delay(2000);

	if (nand_chip()) {
		nand_status = NAND_ABSENT;
		if (mem_type == GPMC_NAND)
			pr_err("Unsupported NAND or no NAND found!\n");
	} else {
		nand_status = NAND_PRESENT;
	}

	return 0;
}

/*********************************************************
 * Routine: nand_init
 * Description: Set up nand for nand and jffs2 commands
 *********************************************************/
int nand_init(void)
{
	if (board->nand_init)
		return board->nand_init();

	return cm_t3x_nand_init(board->nand_gpmc_config);
}

int mmc_boot(unsigned char *buf)
{
	long size = 0;
#ifdef CFG_CMD_FAT
	block_dev_desc_t *dev_desc = NULL;
	unsigned char ret;
	char *u_boot_fname = U_BOOT_FILENAME;

	if (board->u_boot_fname)
		u_boot_fname = board->u_boot_fname;

	ret = mmc_init(1);
	if (ret == 0) {
		pr_err("MMC init failed\n");
		return 0;
	}

	dev_desc = mmc_get_dev(0);
	fat_register_device(dev_desc, 1);
	size = file_fat_read(u_boot_fname, buf, 0);
	if (size == -1) {
		pr_err("Could not read %s from MMC\n", u_boot_fname);
		return 0;
	}

	pr_info("%ld Bytes Read from MMC\n", size);
	pr_info("Starting OS Bootloader from MMC...\n");
#endif
	return size;
}

#define pr_debug_ptr(_ptr, _flag)			\
do {							\
	if (!(_ptr))					\
		pr_debug("NULL\n");			\
	else if ((_flag) == 's')			\
		pr_debug("%s\n", (_ptr));		\
	else if ((_flag) == 'p')			\
		pr_debug("0x%08X\n", (_ptr));		\
} while (0)

/****************************************************************
 * Routine: board_data_dump
 * Description: Dumps a content of the board table and board data
 ****************************************************************/
void board_data_dump(void)
{
	int i;

	pr_debug("\nBoard probe table:\n");
	for (i = 0; i < ARRAY_SIZE(board_probe_table); i++) {
		pr_debug("\tboard_probe_func(%d):\t", i);
		pr_debug_ptr(board_probe_table[i], 'p');
	}

	pr_debug("\nBoard data:\n");
	pr_debug("\tboard_name:\t\t\t");
	pr_debug_ptr(board->board_name, 's');
	pr_debug("\tu_boot_fname:\t\t\t");
	pr_debug_ptr(board->u_boot_fname, 's');
	pr_debug("\tboard_init:\t\t\t");
	pr_debug_ptr(board->board_init, 'p');
	pr_debug("\tmisc_init_r:\t\t\t");
	pr_debug_ptr(board->misc_init_r, 'p');
	pr_debug("\tget_mem_type:\t\t\t");
	pr_debug_ptr(board->get_mem_type, 'p');
	pr_debug("\tnand_init:\t\t\t");
	pr_debug_ptr(board->nand_init, 'p');
	pr_debug("\n");

	pr_debug("\twatchdog_init:\t\t\t");
	pr_debug_ptr(board->watchdog_init, 'p');
	pr_debug("\ttry_unlock_memory:\t\t");
	pr_debug_ptr(board->try_unlock_memory, 'p');
	pr_debug("\tset_muxconf_regs:\t\t");
	pr_debug_ptr(board->set_muxconf_regs, 'p');
	pr_debug("\tprcm_init:\t\t\t");
	pr_debug_ptr(board->prcm_init, 'p');
	pr_debug("\tper_clocks_enable:\t\t");
	pr_debug_ptr(board->per_clocks_enable, 'p');
	pr_debug("\tsetup_memory_controller:\t");
	pr_debug_ptr(board->setup_memory_controller, 'p');
	pr_debug("\n");

	pr_debug("\tcore_dpll_param:\t\t");
	pr_debug_ptr(board->core_dpll_param, 'p');
	pr_debug("\tmpu_dpll_param:\t\t\t");
	pr_debug_ptr(board->mpu_dpll_param, 'p');
	pr_debug("\tiva_dpll_param:\t\t\t");
	pr_debug_ptr(board->iva_dpll_param, 'p');
	pr_debug("\tper_dpll_param:\t\t\t");
	pr_debug_ptr(board->per_dpll_param, 'p');
	pr_debug("\tper_36x_dpll_param:\t\t");
	pr_debug_ptr(board->per_36x_dpll_param, 'p');
	pr_debug("\n");

	pr_debug("\tnand_gpmc_config:\t\t");
	pr_debug_ptr(board->nand_gpmc_config, 'p');
	pr_debug("\n");
}

void board_hang(void)
{
	board_data_dump();
	pr_err("X-Loader hangs\n");

	while (1)
		led_blink(1000, 1);
}
